from clipboard.clipboard import Clipboard
from clipboard.clipboard import ClipboardFormat
from clipboard.clipboard import get_clipboard
from clipboard.clipboard import set_clipboard
from clipboard.html_clipboard import HTML_ENCODING
from clipboard.html_clipboard import HTMLClipboard


__all__ = [
    "Clipboard",
    "ClipboardFormat",
    "HTML_ENCODING",
    "HTMLClipboard",
    "get_clipboard",
    "set_clipboard",
]



import ctypes
from enum import Enum
from enum import EnumMeta
from typing import Any
from typing import List
from typing import Optional

from clipboard._c_interface import CF_HTML
from clipboard._c_interface import CF_RTF
from clipboard._c_interface import EnumClipboardFormats
from clipboard._c_interface import GetClipboardFormatNameA


class ExtendedEnum(EnumMeta):
    def __contains__(cls, item: Any):
        return any(
            [
                item in cls.names,  # type: ignore
                item in cls.values,  # type: ignore
                item in ClipboardFormat.__members__.values(),
            ]
        )


class ClipboardFormat(Enum, metaclass=ExtendedEnum):
    CF_TEXT = 1
    CF_UNICODETEXT = 13
    CF_LOCALE = 16

    CF_HTML = CF_HTML
    CF_RTF = CF_RTF
    HTML_Format = 49418

    text = CF_UNICODETEXT  # alias
    html = HTML_Format  # alias
    HTML = html  # alias
    rtf = CF_RTF  # alias

    @classmethod  # type: ignore
    @property
    def values(cls):
        return [i.value for i in cls]

    @classmethod  # type: ignore
    @property
    def names(cls):
        return [i.name for i in cls]

    def __str__(self):
        return f"{str(self.value)} {str(self.name)}"

    def __eq__(self, other):
        if isinstance(self, type(other)):
            return self.name == other.name and self.value == other.value
        elif isinstance(other, int):
            return self.value == other
        else:
            return False


def get_clipboard_formats(formats: List[int] = None) -> List[int]:
    """Return all available clipboard formats on clipboard.

    First format is the format on the clipboard, depending on your system.
    """
    if formats is None:
        formats = [EnumClipboardFormats(0)]

    last_format: int = formats[-1]
    if last_format == 0:
        return formats[:-1]
    else:
        return formats + [EnumClipboardFormats(last_format)]


def get_format_name(format_code: int) -> Optional[str]:
    """Get the name of the format by it's number.

    C function does not work for standard types (e.g. 1 for CF_TEXT).
    So, this function will use ClipboardFormat for those in the standard.
    """

    # Built-In
    if format_code in ClipboardFormat.values:  # type: ignore
        return ClipboardFormat(format_code).name

    max_buffer_length: int = 100

    buffer: ctypes.Array[ctypes.c_char] = ctypes.create_string_buffer(
        b" " * 200
    )
    return_code: int = GetClipboardFormatNameA(
        format_code, ctypes.byref(buffer), max_buffer_length
    )

    # Failed
    if return_code == 0:
        return None

    format_name: str = buffer.value.decode()

    return format_name



import re
from typing import List
from typing import Optional
from typing import TypeVar


ENCODING = "UTF-8"
HTML_ENCODING = ENCODING


A = TypeVar("A", str, bytes)


class HTMLClipboard:
    """Windows HTML Clipboard"""

    version = 0.9
    template = """
    Version:{version}
    StartHTML:{start_html_byte}
    EndHTML:{end_html_byte}
    StartFragment:{start_fragment_byte}
    EndFragment:{end_fragment_byte}
    SourceURL:{source_url}
    <html>
    <body>
    <!--StartFragment-->
    {fragment}
    <!--EndFragment-->
    </body>
    </html>
    """
    template = "\n".join(
        [i for i in map(str.strip, template.splitlines()) if i]
    )

    def __init__(self, content: str = ""):
        self.fragments: List[str] = []

        self.start_html: int = -1
        self.end_html: int = -1

        self.start_fragment: int = -1
        self.end_fragment: int = -1

        # Optional
        self.start_selection: Optional[str] = None
        self.end_selection: Optional[str] = None

        # WIP
        self.content: str = content
        self.raw: bytes = content.encode(encoding=HTML_ENCODING)

    def generate_template(self) -> str:
        fragments: List[str] = (
            self.fragments if self.fragments else [self.content]
        )

        # Generate Fragments
        result: str = self.generate_fragments(fragments)
        # Generate HTML
        result = self.generate_html(result)
        # Add Header
        result = self.generate_header(result)
        # Get Byte Counts
        result = self.add_byte_counts(result)

        return result

    def generate_fragments(self, fragments: List) -> str:
        results: List[str] = []
        for fragment in fragments:
            results.append("<!--StartFragment-->")
            results.append(f"{fragment}")
            results.append("<!--EndFragment-->")

        # Clean
        result: str = "\n".join(results)

        return result

    def generate_html(self, string: str) -> str:
        lines = string.splitlines()
        body = ["<body>"] + lines + ["</body>"]
        html = ["<html>"] + body + ["</html>"]

        return "\n".join(html)

    def generate_header(self, string: str) -> str:
        lines = string.splitlines()

        version = self.version
        start_html_byte = self.start_html
        end_html_byte = self.end_html
        start_fragment_byte = self.start_fragment
        end_fragment_byte = self.end_fragment
        source_url = None

        if source_url is not None:
            lines.insert(0, f"SourceURL:{source_url}")
        lines.insert(0, f"EndFragment:{end_fragment_byte}")
        lines.insert(0, f"StartFragment:{start_fragment_byte}")
        lines.insert(0, f"EndHTML:{end_html_byte}")
        lines.insert(0, f"StartHTML:{start_html_byte}")
        lines.insert(0, f"Version:{version}")

        return "\n".join(lines)

    def add_byte_counts(self, content: str) -> str:
        # Check
        current_values = self.get_byte_values(content)
        if all((i is not None and i != -1) for i in current_values.values()):
            content = self.update_byte_counts(content)
            return content

        # Setup
        content_bytes: bytes = content.encode(encoding=HTML_ENCODING)

        # Blocks to find
        html_start = "<html>".encode(encoding=HTML_ENCODING)
        html_end = "</html>".encode(encoding=HTML_ENCODING)
        fragment_start = "<!--StartFragment-->".encode(encoding=HTML_ENCODING)
        fragment_end = "<!--EndFragment-->".encode(encoding=HTML_ENCODING)

        # Find Values
        found_html_start = content_bytes.find(html_start)
        found_html_end = content_bytes.find(html_end)
        found_fragment_start = content_bytes.find(fragment_start)
        found_fragment_end = content_bytes.find(fragment_end)

        # Fix Values
        if HTML_ENCODING == "UTF-8":
            found_html_end += len(html_end)
            found_fragment_start += len(fragment_start)

        # Set Values
        self.start_html = found_html_start
        self.end_html = found_html_end
        self.start_fragment = found_fragment_start
        self.end_fragment = found_fragment_end

        # Update
        content_bytes = self.update_byte_counts(content_bytes)

        # Clean Up
        result = content_bytes.decode(encoding=HTML_ENCODING)

        return self.add_byte_counts(result)

    def get_byte_values(self, content: str) -> dict:
        re_StartHTML = re.compile(r"StartHTML:(\d+)", flags=re.MULTILINE)
        StartHTML = int(
            re_StartHTML.findall(content)[0]
            if re_StartHTML.findall(content)
            else -1
        )

        re_EndHTML = re.compile(r"EndHTML:(\d+)", flags=re.MULTILINE)
        EndHTML = int(
            re_EndHTML.findall(content)[0]
            if re_EndHTML.findall(content)
            else -1
        )

        re_StartFragment = re.compile(
            r"StartFragment:(\d+)", flags=re.MULTILINE
        )
        StartFragment = int(
            re_StartFragment.findall(content)[0]
            if re_StartFragment.findall(content)
            else -1
        )

        re_EndFragment = re.compile(r"EndFragment:(\d+)", flags=re.MULTILINE)
        EndFragment = int(
            re_EndFragment.findall(content)[0]
            if re_EndFragment.findall(content)
            else -1
        )

        return {
            "StartHTML": StartHTML,
            "EndHTML": EndHTML,
            "StartFragment": StartFragment,
            "EndFragment": EndFragment,
        }

    def update_byte_counts(self, content: A) -> A:
        data: str
        if isinstance(content, bytes):
            data = content.decode(encoding=HTML_ENCODING)
        elif isinstance(content, str):
            data = content
        else:
            raise TypeError(f"{type(content)} is not a valid type")

        re_value = r"(None|-?\d+)"

        re_StartHTML = re.compile(rf"StartHTML:{re_value}", flags=re.MULTILINE)
        re_EndHTML = re.compile(rf"EndHTML:{re_value}", flags=re.MULTILINE)
        re_StartFragment = re.compile(
            rf"StartFragment:{re_value}", flags=re.MULTILINE
        )
        re_EndFragment = re.compile(
            rf"EndFragment:{re_value}", flags=re.MULTILINE
        )

        data = re.sub(re_StartHTML, rf"StartHTML:{self.start_html}", data)
        data = re.sub(re_EndHTML, rf"EndHTML:{self.end_html}", data)
        data = re.sub(
            re_StartFragment, rf"StartFragment:{self.start_fragment}", data
        )
        data = re.sub(
            re_EndFragment, rf"EndFragment:{self.end_fragment}", data
        )

        if isinstance(content, bytes):
            return data.encode(encoding=HTML_ENCODING)
        elif isinstance(content, str):
            return data
        else:
            raise TypeError(f"{type(content)} is not a valid type")



import ctypes
from typing import List
from typing import Optional
from typing import Union

from clipboard._c_interface import HANDLE
from clipboard._c_interface import LPVOID
from clipboard._c_interface import CloseClipboard
from clipboard._c_interface import EmptyClipboard
from clipboard._c_interface import EnumClipboardFormats
from clipboard._c_interface import GetClipboardData
from clipboard._c_interface import GlobalAlloc
from clipboard._c_interface import GlobalLock
from clipboard._c_interface import GlobalSize
from clipboard._c_interface import GlobalUnlock
from clipboard._c_interface import OpenClipboard
from clipboard._c_interface import SetClipboardData
from clipboard._logging import get_logger
from clipboard.constants import HTML_ENCODING
from clipboard.constants import UTF_ENCODING
from clipboard.formats import ClipboardFormat


hMem = HANDLE  # Type Alias
logger = get_logger(__name__)
GMEM_MOVEABLE = 0x0002
GMEM_ZEROINIT = 0x0040
GMEM_DDESHARE = 0x2000


def get_clipboard(format: Union[int, ClipboardFormat] = None) -> Optional[str]:
    """Conveniency wrapper to get clipboard."""
    with Clipboard() as cb:
        text = cb.get_clipboard(format=format)
    return text


def set_clipboard(
    content: str, format: Union[int, ClipboardFormat] = None
) -> None:
    """Conveniency wrapper to set clipboard."""
    with Clipboard() as cb:
        return cb.set_clipboard(content=content, format=format)


class Clipboard:
    default_format: ClipboardFormat = ClipboardFormat.CF_UNICODETEXT

    def __init__(self, format: Union[ClipboardFormat, str, int] = None):
        if format is None:
            format = self.default_format.value
        else:
            format = self._resolve_format(format)

        self.format: int = format  # type: ignore
        # TODO Add type annotations
        self.h_clip_mem: Optional[HANDLE] = None
        self.address: Optional[LPVOID] = None
        self.size: Optional[int] = None  # in bytes

        self.locked: bool = False
        self.opened: bool = False

    def available_formats(self) -> List[int]:
        """Return all available clipboard formats on clipboard.

        First format is the format on the clipboard, depending on your system.
        """
        # TODO Add this to a base class (HTML clipboard could use this)

        logger.info("Getting available clipboard formats...")

        def get_formats(formats: List = None) -> List[int]:
            if formats is None:
                formats = [EnumClipboardFormats(0)]

            last_format = formats[-1]
            if last_format == 0:
                return formats[:-1]
            else:
                return formats + [EnumClipboardFormats(last_format)]

        available_formats: List[int] = []
        if not self.opened:
            with self:
                available_formats = self.available_formats()
        elif self.opened:
            available_formats = get_formats()

        return available_formats

    def get_clipboard(
        self, format: Union[int, ClipboardFormat] = None
    ) -> Optional[str]:
        """Get data from clipboard, returning None if nothing is on it."""

        if not self.opened:
            with self:
                return self.get_clipboard(format=format)

        if format is None:
            format = self.format
        else:
            format = self._resolve_format(format)

        if format not in self.available_formats():
            formats = self.available_formats()
            message = (
                f"{format} not supported. Choose from following...\n"
                + "\n".join(map(str, formats))
            )
            raise Exception(message)

        # Info
        self.h_clip_mem = GetClipboardData(format)
        self.address = self._lock(self.h_clip_mem)  # type: ignore
        self.size = GlobalSize(self.address)

        if not self.size:
            raise Exception("Get Clipboard failed...")

        string: ctypes.Array[ctypes.c_byte]
        text: str
        if format == ClipboardFormat.CF_UNICODETEXT.value:
            string = (ctypes.c_byte * self.size).from_address(
                int(self.address)  # type: ignore
            )
            text = bytearray(string).decode(encoding=UTF_ENCODING)[:-1]

            return text
        elif (
            format == ClipboardFormat.CF_HTML.value
            or format == ClipboardFormat.HTML_Format.value
        ):
            bytes_ = (ctypes.c_char * self.size).from_address(
                int(self.address)  # type: ignore
            )
            html = bytes(bytes_).decode(HTML_ENCODING)

            return html

        else:
            string = (ctypes.c_byte * self.size).from_address(
                int(self.address)  # type: ignore
            )
            text = bytearray(string).decode(encoding="utf-8")[:-1]

            return text

    def set_clipboard(
        self, content: str, format: Union[int, ClipboardFormat] = None
    ) -> HANDLE:
        """Set clipboard."""

        set_handle: HANDLE = self._set_clipboard(content, format)

        return set_handle

    def _set_clipboard(
        self, content: str, format: Union[int, ClipboardFormat] = None
    ) -> HANDLE:
        """Hides the HANDLE."""

        if format is None:
            format = self.default_format

        if not self.opened:
            with self:
                return self._set_clipboard(content=content, format=format)

        if format is None:
            format = self.format
        format = self._resolve_format(format)
        self._empty()

        set_handle: HANDLE
        alloc_handle: HANDLE
        content_bytes: bytes
        contents_ptr: LPVOID
        if format == ClipboardFormat.CF_UNICODETEXT.value:
            content_bytes = content.encode(encoding="utf-16le")

            alloc_handle = GlobalAlloc(
                GMEM_MOVEABLE | GMEM_ZEROINIT, len(content_bytes) + 2
            )
            contents_ptr = GlobalLock(alloc_handle)
            ctypes.memmove(contents_ptr, content_bytes, len(content_bytes))
            GlobalUnlock(alloc_handle)

            set_handle = SetClipboardData(format, alloc_handle)

        elif (
            format == ClipboardFormat.CF_HTML.value
            or format == ClipboardFormat.HTML_Format.value
        ):
            html_content_bytes: bytes = content.encode(encoding="utf-16le")

            alloc_handle = GlobalAlloc(
                GMEM_MOVEABLE | GMEM_ZEROINIT, len(content) + 2
            )
            contents_ptr = GlobalLock(alloc_handle)  # type: ignore
            ctypes.memmove(contents_ptr, html_content_bytes, len(content))
            GlobalUnlock(alloc_handle)

            set_handle = SetClipboardData(format, alloc_handle)
        else:
            content_bytes = content.encode(encoding="utf-8")

            alloc_handle = GlobalAlloc(GMEM_MOVEABLE, len(content_bytes) + 1)
            contents_ptr = GlobalLock(alloc_handle)
            ctypes.memmove(contents_ptr, content_bytes, len(content_bytes))
            GlobalUnlock(alloc_handle)

            set_handle = SetClipboardData(format, alloc_handle)

        if set_handle is False:
            # handle will be set to NULL
            raise Exception("Set Clipboard failed...")

        return set_handle

    def _resolve_format(self, format: Union[ClipboardFormat, str, int]) -> int:
        """Given an integer, respresenting a clipboard format, or a
        ClipboardFormat object, return the respective integer."""

        if isinstance(format, ClipboardFormat):
            format = format.value
        elif isinstance(format, int):
            pass
        elif isinstance(format, str):
            try:
                format = ClipboardFormat[format].value
            except KeyError:
                # Not a Supported Format
                raise ValueError(
                    f"{format} is not a valid format. Choose from following...\n"
                    + "\n".join(map(str, ClipboardFormat))
                )

        return format  # type: ignore

    def __getitem__(self, format: Union[int, ClipboardFormat] = None):
        return self.get_clipboard(format)

    def __setitem__(self, format, content) -> None:
        format = self._resolve_format(format)
        self._empty()
        self.set_clipboard(content, format)

    def __enter__(self):
        if self._open():
            return self
        else:
            raise Exception("Unable to open clipboard.")

    def __exit__(
        self, exception_type, exception_value, exception_traceback
    ) -> bool:
        if exception_type is not None:
            import traceback

            traceback.print_exception(
                exception_type, exception_value, exception_traceback
            )

        self._close()
        return True

    def _open(self, handle: int = None) -> bool:
        self.opened = True
        return OpenClipboard(handle)

    def _close(self) -> bool:
        self.opened = False
        self._unlock()
        return CloseClipboard()

    def _lock(self, handle: HANDLE) -> LPVOID:
        self.locked = True
        # if fails, GlobalLock returns NULL (False in Python)
        return GlobalLock(handle)

    def _unlock(self, handle: HANDLE = None) -> int:
        if handle is None:
            handle = self.h_clip_mem

        self.locked = False
        return GlobalUnlock(handle)

    def _empty(self) -> int:
        if not self.opened:
            with self:
                return self._empty()
        elif self.opened:
            return_code = EmptyClipboard()
        else:
            raise RuntimeError("Failed to empty clipboard.")

        return return_code



""" Helper to interface with the C code.

Note: This did not feel like a simple process to identify the paths to access
the C functions.
"""

import ctypes
from ctypes.wintypes import BOOL
from ctypes.wintypes import HANDLE
from ctypes.wintypes import HGLOBAL
from ctypes.wintypes import HWND
from ctypes.wintypes import LPSTR
from ctypes.wintypes import LPVOID
from ctypes.wintypes import LPWSTR
from ctypes.wintypes import UINT


# C Libraries
windll = ctypes.windll  # type: ignore
user32 = windll.user32
kernel32 = windll.kernel32

# C Functions
OpenClipboard = user32.OpenClipboard
OpenClipboard.argtypes = [HWND]
OpenClipboard.restype = BOOL

CloseClipboard = user32.CloseClipboard
CloseClipboard.argtypes = []
CloseClipboard.restype = BOOL

SetClipboardData = user32.SetClipboardData
SetClipboardData.argtypes = [UINT, HANDLE]
SetClipboardData.restype = HANDLE

EmptyClipboard = user32.EmptyClipboard
EmptyClipboard.argtypes = []
EmptyClipboard.restype = BOOL

GetClipboardData = user32.GetClipboardData
GetClipboardData.argtypes = [UINT]
GetClipboardData.restype = HANDLE


IsClipboardFormatAvailable = user32.IsClipboardFormatAvailable
IsClipboardFormatAvailable.argtypes = [UINT]
IsClipboardFormatAvailable.restype = BOOL

# Returns first available clipboard format in a specified list
GetPriorityClipboardFormat = user32.GetPriorityClipboardFormat
GetPriorityClipboardFormat.argtypes = [UINT, ctypes.c_int]
GetPriorityClipboardFormat.restype = ctypes.c_int

# w - unicode (utf-16 on windows)
# a - ansi
GetClipboardFormatNameA = user32.GetClipboardFormatNameA
GetClipboardFormatNameA.argtypes = [UINT, LPSTR, ctypes.c_int]
GetClipboardFormatNameA.restype = ctypes.c_int

GetClipboardFormatNameW = user32.GetClipboardFormatNameW
GetClipboardFormatNameW.argtypes = [UINT, LPWSTR, ctypes.c_int]
GetClipboardFormatNameW.restype = ctypes.c_int

GlobalLock = kernel32.GlobalLock
GlobalLock.argtypes = [HGLOBAL]
GlobalLock.restype = LPVOID

GlobalUnlock = kernel32.GlobalUnlock
GlobalUnlock.argtypes = [HGLOBAL]
GlobalUnlock.restype = BOOL

GlobalSize = kernel32.GlobalSize
GlobalSize.argtypes = [HGLOBAL]
GlobalSize.restype = int  # SIZE_T

GlobalAlloc = kernel32.GlobalAlloc
GlobalAlloc.argtypes = [UINT, ctypes.c_int]
GlobalAlloc.restype = HANDLE


EnumClipboardFormats = user32.EnumClipboardFormats
EnumClipboardFormats.argtypes = [UINT]
EnumClipboardFormats.restype = UINT


# Additional
CF_HTML = ctypes.windll.user32.RegisterClipboardFormatW("HTML Format")  # type: ignore
CF_RTF = ctypes.windll.user32.RegisterClipboardFormatW("Rich Text Format")  # type: ignore



import setuptools


with open("README.md", "r", encoding="utf-8") as fin:
    long_description = fin.read()


build_requires = [
    "wheel",
    "setuptools",
]
test_requires = [
    "tox",
    "tox-gh-actions",
]
dev_requires = (
    build_requires
    + test_requires
    + [
        # CI/CD Tool - Runs formatting tools
        "pre-commit",
        # Type Checker
        "mypy",
    ]
)


setuptools.setup(
    name="clip-util",
    version="0.1.17",
    author="Kyle L. Davis",
    author_email="aceofspades5757.github@gmail.com",
    url="https://github.com/AceofSpades5757/clip-util",
    project_urls={
        "Documentation": "https://clip-util.readthedocs.io/en/latest/",
        "Author": "https://github.com/AceofSpades5757",
    },
    description="Clipboard utilities for use with Python.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    python_requires=">=3.9",
    packages=["clipboard"],
    package_dir={"": "src"},
    extras_require={
        "dev": dev_requires,
        "build": build_requires,
        "test": test_requires,
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: Microsoft :: Windows",
        "License :: OSI Approved :: MIT License",
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
)



import logging.config
from pathlib import Path


logs_path = Path(__file__).parent / "logs"
logs_path.mkdir(exist_ok=True)


date_format: str = "%Y-%m-%dT%H:%M:%S%z"
message_format = "{asctime} - {name} - {levelname:<8} - {message}"
logging_config = dict(
    version=1,
    disable_existing_loggers=False,
    root=dict(
        level=logging.INFO,
        handlers=[
            "file_handler",
        ],
    ),
    formatters={
        "default": {
            # 'date_format': '%Y-%m-%dT%H:%M:%S+0000%z',
            "date_format": date_format,
            "format": message_format,
            "style": "{",
        },
    },
    handlers={
        "file_handler": {
            "level": logging.INFO,
            "class": "logging.FileHandler",
            "formatter": "default",
            "filename": logs_path / "root.log",
        },
    },
    loggers={},
)

logging.config.dictConfig(logging_config)
formatter: logging.Formatter = logging.Formatter(
    datefmt=date_format,
    style="{",
    fmt=message_format,
)
logging.getLogger().handlers[0].setFormatter(formatter)


def get_logger(
    name: str, level=logging.DEBUG, debug: bool = False
) -> logging.Logger:
    logger: logging.Logger = logging.getLogger(__name__)
    logger.setLevel(level)

    log_file: Path = logs_path / Path(name).with_suffix(".log").name

    formatter: logging.Formatter = logging.Formatter(
        datefmt=date_format,
        style="{",
        fmt=message_format,
    )

    file_handler: logging.FileHandler = logging.FileHandler(log_file)
    file_handler.setLevel(level)
    file_handler.setFormatter(formatter)

    logger.addHandler(file_handler)

    if debug:
        stream_handler: logging.StreamHandler = logging.StreamHandler()
        stream_handler.setLevel(logging.DEBUG)
        logger.addHandler(stream_handler)

    return logger
